Parser for Featherweight Java
(C) Andreas Abel 2002

> {
> {-# OPTIONS -fno-warn-overlapping-patterns #-}
>
> module FFJ.Parser where
>
> import Char;
> }


First thing to declare is the name of your parser,
and the type of the tokens the parser reads.

> %name parser
> %tokentype { Token }


The parser will be of type [Token] -> ?, where ? is determined by the
production rules.  Now we declare all the possible tokens:

> %token 
>       class           { TokenClass }
>       refines         { TokenRefines }
>       extends         { TokenExtends }
>       override        { TokenOverride }
>       object          { TokenObject }
>       this            { TokenThis }
>       new             { TokenNew }
>       return          { TokenReturn }
>	id		{ TokenID $$ }
>       '.'             { TokenDot }
>       ','             { TokenComma }
>       ';'             { TokenSemicolon }
>	'='		{ TokenEq }
>	'('		{ TokenLParen }
>	')'		{ TokenRParen }
>       '{'             { TokenLBrace }
>       '}'             { TokenRBrace }


Like yacc, we include %% here, for no real reason.

> %%

Statements.


> Stats :: { [PStat] }
> Stats : Stats1                    { reverse $1 }
>
> Stats1 :: { [PStat] }
> Stats1 : Stats1 Class             { $2 : $1}
>        | Stats1 Ref               { $2 : $1}
>        | Stats1 Def               { $2 : $1}
>        |                          { [] }

Class production rules.


A class identifier CID is either "Object" or an identifier. As Java, we permit both upper and lower case identifiers for classes.

> CID   :: { PCID }
> CID   : object                    { PObject }
>       | id                        { PCID $1 }

Since both field and constructor declarations start with an identifier,
we need cannot decide right from the start which of these declarations
we are parsing.  We need to see another token.  Is it an identifier,
then we are parsing a field declaration, is it an opening parenthesis,
we are parsing the constructor definition.  Hence, we introduce the
non-terminal FC which denotes "a possibly empty list of field
declarations plus finally the constructor code".

> Class :: { PStat }
> Class : class id extends CID '{' FieldsAndMethods '}'   { PClass $2 $4 (fst $6) (snd $6) } 
>
> Ref :: { PStat }
> Ref : refines class id '{' FieldsAndMethods '}'   { PRef $3 (fst $5) (snd $5) } 


Decls is a non empty list of "CID ID" declarations as used for
constructor and method argument lists.  Declo is a possibly empty list
of such declarations (the "o" for "optional).

> Declo :: { [PDecl] }
> Declo :                           { [] }
>       | Decls                     { reverse $1 }
>
> Decls :: { [PDecl] }
> Decls : Decls ',' CID id          { PDecl $3 $4 : $1 }
>       | CID id                    { [PDecl $1 $2] }

Method definitions.

    Methods :
            | CID id (Declo) { return Term; } Methods

> FieldsAndMethods :: { ([PField],[PMethod]) }
> FieldsAndMethods : FieldOrMethod1                 { (reverse [f|Left f<- $1], reverse [m|Right m<-$1]) }
>
> FieldOrMethod1 :: { [Either PField PMethod] }
> FieldOrMethod1 : FieldOrMethod1 FieldOrMethod          { $2 : $1 }
>         |                         { [] }
>
> FieldOrMethod :: { Either PField PMethod }
> FieldOrMethod : override CID id '(' Declo ')' '{' return Term ';' '}'  { Right (PMethod True (PDecl $2 $3) $5 $9) } 
>        | CID id '(' Declo ')' '{' return Term ';' '}'  { Right (PMethod False (PDecl $1 $2) $4 $8) } 
>        | CID id ';'             { Left (PField (PDecl $1 $2))}

Term production rules.

    Term : Var              Variable 'this' or id
         | Term Elims       Elimination: field lookup or method invokation
         | (CID) Term       Cast
         | (Term)           Parentheses to group terms and resolve ambiguities
         | new CID (Terms)  Object construction

The main issue here is to resolve the ambiguity between terms in
parentheses (Term) and casts (CID) Term.  Since there is no reason to
put a single variable in parentheses, we interprete "(ID)" as cast.
In practise, we need to distiguish terms which consist of a single
variable from more complex terms (Term1).

> Term  :: { PTerm }
> Term  : Var                       { PVar $1 }
>       | Term1                     { $1 }
>
> Term1 :: { PTerm }
> Term1 : Var '.' Elim Elims        { PElim (PVar $1) ($3 : $4) }
>       | '(' CID ')' Term          { PCast $2 $4}
>       | '(' Term1 ')' Elims       { PElim $2 $4 }  
>       | new CID '(' Termo ')'     { PNew $2 $4 }
>
> Var   :: { PVar }
> Var   : this                      { PThis }
>       | id                        { PVID $1 }

An elimination Elim is either a field lookup ("Get") or a method
invocation ("Call").

> Elims :: { [PElim] }
> Elims : Elims1                    { reverse $1 }
>
> Elims1:: { [PElim] }
> Elims1: Elims '.' Elim            { $3 : $1 }
>       |                           { [] }
>
> Elim  :: { PElim }
> Elim  : id '(' Termo ')'          { PCall $1 $3 }
>       | id                        { PGet $1 }

Term lists.

> Termo :: { [PTerm] }
> Termo :                           { [] }
>       | Terms                     { reverse $1 }
>
> Terms :: { [PTerm] }
> Terms : Terms ',' Term            { $3 : $1 }
>       | Term                      { [$1] }

Variable definitions. 

> Def   :: { PStat }
> Def   : CID id '=' Term ';'       { PDef $1 $2 $4 }

> {

Pre-terms as parser result.

> data PVar    = PThis 
>       | PVID String
>	 deriving Show
> data PCID    = PObject 
>       | PCID String
>	 deriving Show
> data PTerm   = PVar PVar
>       | PElim PTerm [PElim]
>       | PCast PCID PTerm
>       | PNew PCID [PTerm]
>	 deriving Show
> data PElim   = PGet String
>       | PCall String [PTerm]
>	 deriving Show

Pre-statements.

> data PDecl   = PDecl PCID String
>	 deriving Show
> data PMethod = PMethod Bool PDecl [PDecl] PTerm
>	 deriving Show
> data PField = PField PDecl
>	 deriving Show
> data PStat   = PClass String PCID [PField] [PMethod]
>		| PRef String [PField] [PMethod]
>       | PDef PCID String PTerm 
>	 deriving Show

Parser error.

All parsers must declair this function, which is called when an error
is detected.  Note that currently we do no error recovery.

> happyError :: [Token] -> a
> happyError toks = error ("Parse error" ++ show toks ++ "\n")

The lexer.

> data Token = TokenClass
>     | TokenRefines
>     | TokenExtends
>     | TokenOverride
>     | TokenObject
>     | TokenThis
>     | TokenNew
>     | TokenReturn
>     | TokenID String
>     | TokenDot
>     | TokenComma
>     | TokenSemicolon
>     | TokenEq
>     | TokenLParen
>     | TokenRParen
>     | TokenLBrace
>     | TokenRBrace
>       deriving Show

> lexer :: String -> [Token]
> lexer [] = []
> lexer (c:cs) 
>	| isSpace c = lexer cs
> 	| isAlpha c = lexID (c:cs)
> lexer ('.':cs) = TokenDot       : lexer cs
> lexer (',':cs) = TokenComma     : lexer cs
> lexer (';':cs) = TokenSemicolon : lexer cs
> lexer ('=':cs) = TokenEq        : lexer cs
> lexer ('(':cs) = TokenLParen    : lexer cs
> lexer (')':cs) = TokenRParen    : lexer cs
> lexer ('{':cs) = TokenLBrace    : lexer cs
> lexer ('}':cs) = TokenRBrace    : lexer cs
> lexer ('/':'/':cs) = lexer (dropWhile (not . (==) '\n') cs)
> lexer ('/':'*':cs) = lexComment cs

> lexComment [] = error "unclosed comment at end of input"
> lexComment ('*':'/':cs) = lexer cs
> lexComment (_:cs) = lexComment cs

> lexID cs =
>    case span isAlphaNum cs of
>	("class",   rest) -> TokenClass   : lexer rest
>	("refines", rest) -> TokenRefines : lexer rest
>	("extends", rest) -> TokenExtends : lexer rest
>	("override", rest) -> TokenOverride : lexer rest
>	("Object",  rest) -> TokenObject  : lexer rest
>	("this",    rest) -> TokenThis    : lexer rest
>	("return",  rest) -> TokenReturn  : lexer rest
>	("new",     rest) -> TokenNew     : lexer rest
>	(id,        rest) -> TokenID id   : lexer rest

main = print . parser . lexer
>
>
End of file.

> }
